@doc raw"""
    (z, cz, fz, status, Δ, Δℓ, γ, δ) = tangent_step!(nlp, z, λ, cz, normcz, fz, LDL, vals, g, ℓzλ, gBg, ρ, γ, δ, meta, workspace; kwargs...)

Solve the following problem
```math
\begin{aligned}
\min_{d} \quad & q(d):=\frac{1}{2} d^T B d + d^T g \\
\text{s.t.} \quad & Ad = 0,
& \|d\| \leq \Delta,
\end{aligned}
```
where `B` is an approximation of the Hessian of the Lagrangian, `A` is the jacobian
matrix, and `g` is the projected gradient.

# Output
- `z`, `cz`, and `fz`: the new iterate, and updated evaluations.
- `status`: computation status.
- `Δ`: trust-region parameter.
- `Δℓ`: differential in Lagrangian computation.
- `γ`, `δ`: updated values for the regularization parameters.

Return `status` with possible outcomes:
- `:cauchy_step`, `:newton`, `:dogleg` if successful step.
- `:unknown` if we didn't enter the loop.
- `:small_horizontal_step`.
- `:tired` if we stop due to `max_eval` or `max_time`.
- `:success` if we computed `z` such that `‖c(z)‖ ≤ meta.ρbar * ρ` and `Δℓ ≥ η₁ q(d)`.

See [`SolverTools.jl`](https://github.com/JuliaSmoothOptimizers/SolverTools.jl/blob/78f6793f161c3aac2234aee8a27aa07f1df3e8ee/src/trust-region/trust-region.jl#L37)
for `SolverTools.aredpred`
"""
function tangent_step!(
  nlp::AbstractNLPModel,
  z::AbstractVector{T},
  λ::AbstractVector{T},
  cz::AbstractVector{T},
  normcz::T,
  fz::T,
  LDL::SymCOOSolver,
  vals::AbstractVector{T},
  g::AbstractVector{T},
  ℓzλ::T,
  gBg::T,
  ρ::AbstractFloat,
  γ::T,
  δ::T,
  meta::MetaDCI,
  workspace::DCIWorkspace,
  verbose::Bool;
  Δ::AbstractFloat = meta.tan_Δ,
  η₁::AbstractFloat = meta.tan_η₁,
  η₂::AbstractFloat = meta.tan_η₂,
  σ₁::AbstractFloat = meta.tan_σ₁,
  σ₂::AbstractFloat = meta.tan_σ₂,
  small_d::AbstractFloat = meta.tan_small_d,
  max_eval::Int = 1_000,
  max_time::AbstractFloat = 1_000.0,
) where {T}
  d, tr, xt = workspace.dtan, workspace.tr, workspace.xtan
  Δℓ = zero(T)
  r = -one(T)

  status = :unknown
  iter = 0
  start_time = time()
  el_time = 0.0

  tired = neval_obj(nlp) + neval_cons(nlp) > max_eval || el_time > max_time

  while !((normcz ≤ meta.ρbar * ρ && r ≥ η₁) || tired)
    #Compute a descent direction d (no evals)
    d, dBd, status, γ, δ, vals =
      compute_descent_direction!(nlp, gBg, g, Δ, LDL, γ, δ, vals, d, meta, workspace, verbose)
    n2d = dot(d, d)
    if √n2d > Δ
      d .*= Δ / √n2d #Just in case.
    end
    if √n2d < small_d
      status = :small_horizontal_step
      break
    end

    @. xt = z + d
    cons_norhs!(nlp, xt, cz)
    normcz = norm(cz)

    if normcz ≤ meta.ρbar * ρ
      ft = obj(nlp, xt)
      ℓxtλ = ft + dot(λ, cz)
      qd = dBd / 2 + dot(g, d)

      Δℓ, pred = aredpred!(tr, nlp, ℓzλ, ℓxtλ, qd, xt, d, dot(g, d))

      r = Δℓ / qd
      if r < η₁ #we decrease further Δ so that ≤ ||d||
        m = max(ceil(log(√n2d / Δ) / log(σ₁)), 1)
        Δ *= σ₁^m
      else #success
        status = :success
        z .= xt
        fz = ft
        ℓzλ = ℓxtλ
        if r ≥ η₂ && √n2d ≥ 0.99 * Δ
          Δ *= σ₂
        end
      end
    else
      m = max(ceil(log(√n2d / Δ) / log(σ₁)), 1)
      Δ *= σ₁^m
    end

    el_time = time() - start_time
    verbose && @info log_row(
      Any[
        "Tr",
        iter,
        neval_obj(nlp) + neval_cons(nlp),
        fz,
        ℓzλ,
        Float64,
        normcz,
        Float64,
        Float64,
        status,
        √n2d,
        Δ,
        el_time,
      ],
    )
    iter += 1

    tired = neval_obj(nlp) + neval_cons(nlp) > max_eval || el_time > max_time
  end

  if tired
    status = :tired
  end

  return z, cz, fz, status, Δ, Δℓ, γ, δ #ℓzλ
end

@doc raw"""
    (d, dBd, status, γ, δ, vals) = compute_descent_direction!(nlp::AbstractNLPModel, gBg, g, Δ, LDL, γ, δ, vals, d, meta, workspace)

Compute a direction `d` solution of
```math
\begin{aligned}
\min_{d} \quad & q(d) \\
\text{s.t.} \quad & \|d\| \leq \Delta.
\end{aligned}
```

# Output
The returned arguments are:
- `d`: the computed direction.
- `dBd`: updated scalar product.
- `status`: computation status.
- `γ`, `δ`: updated values for the regularization parameters.
- `vals`: update values for the SQD system.

Return `status` has four possible outcomes:
- `:cauchy_step`
- `:newton`
- `:dogleg`
- `:interior_cauchy_step` when γ is too large.
"""
function compute_descent_direction!(
  nlp::AbstractNLPModel,
  gBg::T,
  g::AbstractVector{T},
  Δ::T,
  LDL::SymCOOSolver,
  γ::T,
  δ::T,
  vals::AbstractVector{T},
  d::AbstractVector{T},
  meta::MetaDCI,
  workspace::DCIWorkspace,
  verbose::Bool,
) where {T}
  m, n = nlp.meta.ncon, nlp.meta.nvar
  dcp = workspace.dcp

  #first compute a gradient step
  dcp_on_boundary, dcp, dcpBdcp, α = _compute_gradient_step!(nlp, gBg, g, Δ, dcp)

  if dcp_on_boundary # When the Cauchy step is in the boundary, we use it
    status = :cauchy_step
    dBd = dcpBdcp
    d = dcp
  else
    dn, dnBdn, dcpBdn, γ_too_large, γ, δ, vals =
      _compute_newton_step!(nlp, LDL, g, γ, δ, dcp, vals, meta, workspace, verbose)
    norm2dn = dot(dn, dn)
    if γ_too_large || dnBdn ≤ 1e-10 #or same test as gBg in _compute_gradient_step ?
      #dn = 0 here.
      if norm(dcp) < Δ #just to be sure
        d = dcp
        dBd = dcpBdcp
        status = :interior_cauchy
        return d, dBd, status, γ, δ, vals
      end
    end

    if √norm2dn < Δ # Both Newton and Cauchy are inside the TR.
      status = :newton
      dBd = dnBdn
      d = dn
    else
      dotdndcp, norm2dcp = dot(dn, dcp), dot(dcp, dcp)
      τ = _compute_step_length(norm2dn, dotdndcp, norm2dcp, Δ)
      @. d = dn + τ * (dcp - dn)
      dBd = τ^2 * dcpBdcp + 2 * τ * (1 - τ) * dcpBdn + (1 - τ)^2 * dnBdn
      status = :dogleg
    end
  end

  return d, dBd, status, γ, δ, vals
end

@doc raw"""
    _compute_gradient_step!(nlp, gBg, g, Δ, dcp)

Solve the following problem
```math
\begin{aligned}
\min_{α} \quad & q(-α g) \\
\text{s.t.} \quad & \|αg\| \leq \Delta.
\end{aligned}
```

# Output
The returned arguments are:
- `dcp_on_boundary` true if `‖αg‖ = Δ`,
- `dcp = - α g`,
- `dcpBdcp = α^2 gBg`,
- `α` the solution.
"""
function _compute_gradient_step!(
  nlp::AbstractNLPModel,
  gBg::T,
  g::AbstractVector{T},
  Δ::T,
  dcp::AbstractVector{T},
) where {T}
  dcp_on_boundary = false
  dgg = dot(g, g)
  if gBg ≤ 1e-12 * dgg #generalize this test
    α = Δ / √dgg #norm(g)
    dcp_on_boundary = true
  else
    α = dgg / gBg #dot(g, g) / gBg
    if α > Δ / √dgg #norm(g)
      α = Δ / √dgg #norm(g)
      dcp_on_boundary = true
    end
  end
  @. dcp = -α * g
  dcpBdcp = α^2 * gBg

  return dcp_on_boundary, dcp, dcpBdcp, α
end

"""
    _compute_step_length(norm2dn, dotdndcp, norm2dcp, Δ)

Given two directions `dcp` and `dn`, compute the largest `0 ≤ τ ≤ 1` such that `‖dn + τ (dcp -dn)‖ = Δ`.
Returns `τ`.
"""
function _compute_step_length(norm2dn::T, dotdndcp::T, norm2dcp::T, Δ::T) where {T <: AbstractFloat}
  # d = τ dcp + (1 - τ) * dn = dn + τ * (dcp - dn)
  # ‖d‖² = Δ² => τ² ‖dcp - dn‖² + 2τ dnᵀ(dcp - dn) + ‖dn‖² - Δ² = 0
  # Δ = b² - 4ac
  scal = norm2dcp - 2 * dotdndcp + norm2dn
  q₀ = (norm2dn - Δ^2) / scal
  q₁ = 2 * (dotdndcp - norm2dn) / scal
  q₂ = one(T)
  # q₀, q₁, q₂ = [q₀, q₁, q₂] / q₂ #so the first coefficient is 1.
  roots = Krylov.roots_quadratic(q₂, q₁, q₀)
  τ = length(roots) == 0 ? one(T) : min(one(T), roots...)
  return τ
end

include("factorization.jl")
#=
dn, dnBdn, dcpBdn,
γ_too_large, γ, δ, vals = _compute_newton_step!(nlp, LDL, g, γ, δ, dcp, vals)
=#
